from experiment_design.orthogonal_sampling import OrthogonalSamplingDesigner
from experiment_design.random_sampling import RandomSamplingDesigner
from experiment_design.variable import (
    ParameterSpace,
    create_continuous_uniform_space,
    create_discrete_uniform_space,
)

__version__ = "v0.1.1"

__all__ = [
    "OrthogonalSamplingDesigner",
    "RandomSamplingDesigner",
    "ParameterSpace",
    "create_discrete_uniform_space",
    "create_continuous_uniform_space",
]



import numpy as np
from scipy.linalg import solve_triangular


def iman_connover_transformation(
    doe: np.ndarray,
    target_correlation: np.ndarray,
    means: np.ndarray | None = None,
    standard_deviations: np.ndarray | None = None,
) -> np.ndarray:
    """
    Rearrange the values of doe to reduce correlation error while adhering to any marginal constraints of the values
    such as an |LHS|

    :param doe: Array with shape (n_sample, n_dim) representing the initial |DoE| with arbitrary
        correlation.
    :param target_correlation: Symmetric positive definite correlation matrix with shape (n_dim, n_dim) representing
        the desired correlation between variables
    :param means: Array with shape (n_dim,) representing the means of the marginal distributions. If None, it will be
        inferred from doe
    :param standard_deviations: Array with shape (n_dim,) representing the standard deviations of the marginal
        distributions. If None, it will be inferred from doe
    :return: New |DoE| with the same shape and values as doe but smaller correlation error wrt. target_correlation

    References
    ----------
    R.L. Iman and W.J. Conover (1982). “`A distribution-free approach to inducing rank correlation among input variables
    <https://www.researchgate.net/publication/243048186_A_Distribution-Free_Approach_to_Inducing_Rank_Correlation_Among_Input_Variates>`_”


    C. Bogoclu (2022). "`Local Latin Hypercube Refinement for Uncertainty Quantification and Optimization
    <https://hss-opus.ub.ruhr-uni-bochum.de/opus4/frontdoor/deliver/index/docId/9143/file/diss.pdf>`_" Chapter 4.3.2

    Examples
    --------
    >>> from experiment_design.covariance_modification import iman_connover_transformation
    >>> import numpy as np
    >>> from scipy import stats
    >>> np.random.seed(1337)
    >>> samples = stats.randint(0, 100).rvs((30, 2))
    >>> correlation_error = np.max(np.abs(np.corrcoef(samples, rowvar=False) - np.eye(2)))
    >>> new_samples = iman_connover_transformation(samples, np.eye(2))
    >>> np.max(np.abs(np.corrcoef(new_samples, rowvar=False) - np.eye(2))) < correlation_error
    True
    >>> sorted(samples[:, 0]) == sorted(new_samples[:, 0])
    True
    >>> sorted(samples[:, 1]) == sorted(new_samples[:, 1])
    True
    """
    transformed = second_moment_transformation(
        doe, target_correlation, means, standard_deviations
    )
    order = np.argsort(np.argsort(transformed, axis=0), axis=0)
    return np.take_along_axis(np.sort(doe, axis=0), order, axis=0)


def second_moment_transformation(
    doe: np.ndarray,
    target_correlation: np.ndarray,
    means: np.ndarray | None = None,
    standard_deviations: np.ndarray | None = None,
    jitter: float = 1e-6,
) -> np.ndarray:
    """
    Second-moment transformation for achieving the target covariance

    :param doe: Array with shape (n_sample, n_dim) representing the initial design of experiment with arbitrary
        correlation.
    :param target_correlation: Symmetric positive definite correlation matrix with shape (n_dim, n_dim) representing
        the desired correlation between variables
    :param means: Array with shape (n_dim,) representing the means of the marginal distributions. If None, it will be
        inferred from doe
    :param standard_deviations: Array with shape (n_dim,) representing the standard deviations of the marginal
        distributions. If None, it will be inferred from doe
    :param jitter: A small positive constant that will be added to the diagonal of the covariance matrix in case
        it is positive semi-definite to enable Cholesky decomposition.
    :return: New |DoE| with the same shape but different values as doe, that matches the target_correlation exactly.


    Examples
    --------
    >>> from experiment_design.covariance_modification import iman_connover_transformation
    >>> import numpy as np
    >>> from scipy import stats
    >>> np.random.seed(1337)
    >>> samples = stats.norm.rvs(size=(50, 2))
    >>> new_samples = iman_connover_transformation(samples, np.eye(2))
    >>> correlation_error = np.max(np.abs(np.corrcoef(new_samples, rowvar=False) - np.eye(2)))
    >>> bool(np.isclose(correlation_error, 0, atol=1e-6))
    True

    """
    if means is None:
        means = np.mean(doe, axis=0)
    if standard_deviations is None:
        standard_deviations = np.std(doe, axis=0, keepdims=True)
        standard_deviations = standard_deviations.reshape((1, -1))
    target_covariance = (
        standard_deviations.T.dot(standard_deviations) * target_correlation
    )  # convert to covariance before Cholesky
    try:
        target_cov_upper = np.linalg.cholesky(target_covariance).T
    except np.linalg.LinAlgError:
        target_cov_upper = np.linalg.cholesky(
            target_covariance + np.eye(target_covariance.shape[0]) * jitter
        ).T
    cur_cov = np.cov(doe, rowvar=False)
    try:
        cur_cov_upper = np.linalg.cholesky(cur_cov).T
    except np.linalg.LinAlgError:
        cur_cov_upper = np.linalg.cholesky(
            cur_cov + np.eye(cur_cov.shape[0]) * jitter
        ).T

    inv_cov_upper = solve_triangular(cur_cov_upper, target_cov_upper)
    return (doe - means).dot(inv_cov_upper) + means



import logging
from functools import partial
from typing import Callable

import numpy as np
from scipy.stats import uniform

from experiment_design.covariance_modification import iman_connover_transformation
from experiment_design.experiment_designer import ExperimentDesigner
from experiment_design.optimize import (
    random_search,
    simulated_annealing_by_perturbation,
)
from experiment_design.scorers import Scorer, ScorerFactory, select_local
from experiment_design.variable import ParameterSpace


class OrthogonalSamplingDesigner(ExperimentDesigner):
    """
    Create or extend an orthogonal sampling design. Orthogonal sampling design partitions the design space into
    bins of equal marginal probability and places samples such that each bin is only filled once for each dimension.
    If all variables are uniform, orthogonal sampling becomes an |LHS|.

    :param inter_bin_randomness: Controls the randomness of placed points between the bin bounds. Specifically, 0 means
        the points are placed at the center of each bin, whereas 1 leads to a random point placement within the bounds.
        Any other fractions leads to a random placement within that fraction of the bin bounds in each dimension.
    :param non_occupied_bins: Only relevant for extending the design, i.e. if old points are provided, and if the constraint
        regarding the number of occupation of each bin has to be violated. True means that each bin is occupied at least
        once for each dimension, although some bins might be occupied more often. Otherwise, each bin is occupied once
        or less often, leading to empty bins in some cases.
    :param scorer_factory: A factory that creates scorers for the given variables, sample_size and in the cast of an
        extension, old sampling points. If not passed, a default one will be created, that evaluates the maximum
        correlation error and minimum pairwise distance. See
        `experiment_design.scorers.create_default_scorer_factory <#experiment_design.scorers.create_default_scorer_factory>`_
        for more details.



    References
    ----------
    M.D. McKay, W.J. Conover and R.J. Beckmann (1979). “`A comparison of three methods for selecting values of input
    variables in the analysis of output from a computer code
    <https://www.researchgate.net/publication/235709905_A_Comparison_of_Three_Methods_for_Selecting_Vales_of_Input_Variables_in_the_Analysis_of_Output_From_a_Computer_Code>`_”

    A.B. Owen (1992). “`Orthogonal arrays for computer experiments, integration and visualization
    <https://www3.stat.sinica.edu.tw/statistica/oldpdf/A18n17.pdf>`_”

    C. Bogoclu (2022). "`Local Latin Hypercube Refinement for Uncertainty Quantification and Optimization
    <https://hss-opus.ub.ruhr-uni-bochum.de/opus4/frontdoor/deliver/index/docId/9143/file/diss.pdf>`_"
    Chapters 4.3.1 and 5


    Examples
    --------
    >>> from experiment_design import create_continuous_uniform_space, OrthogonalSamplingDesigner
    >>> space = create_continuous_uniform_space([-2., -2.], [2., 2.])
    >>> designer = OrthogonalSamplingDesigner()
    >>> doe1 = designer.design(space, sample_size=20)
    >>> doe1.shape
    (20, 2)
    >>> doe2 = designer.design(space, sample_size=4, old_sample=doe1)
    >>> doe2.shape
    (4, 2)

    """

    def __init__(
        self,
        inter_bin_randomness: float = 0.8,
        non_occupied_bins: bool = False,
        scorer_factory: ScorerFactory | None = None,
    ) -> None:
        self.inter_bin_randomness = inter_bin_randomness
        if non_occupied_bins:
            self.empty_size_check = np.min
        else:
            self.empty_size_check = np.max
        super(OrthogonalSamplingDesigner, self).__init__(scorer_factory=scorer_factory)

    def _create(
        self,
        space: ParameterSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
    ) -> np.ndarray:
        if initial_steps + final_steps <= 1:
            return create_orthogonal_design(
                space=space,
                sample_size=sample_size,
                inter_bin_randomness=self.inter_bin_randomness,
            )

        logging.info("Creating an initial design...")
        doe = random_search(
            creator=partial(
                create_orthogonal_design,
                space=space,
                sample_size=sample_size,
                inter_bin_randomness=self.inter_bin_randomness,
            ),
            scorer=scorer,
            steps=initial_steps,
        )
        logging.info("Optimizing the initial design...")
        return simulated_annealing_by_perturbation(doe, scorer, steps=final_steps)

    def _extend(
        self,
        old_sample: np.ndarray,
        space: ParameterSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
    ) -> np.ndarray:
        local_doe = select_local(old_sample, space)
        probabilities = space.cdf_of(local_doe)
        if not np.all(np.isfinite(probabilities)):
            raise RuntimeError(
                "Non-finite probability encountered. Please check the distributions."
            )

        bins_per_dimension = sample_size + local_doe.shape[0]

        empty = _find_sufficient_empty_bins(
            probabilities,
            bins_per_dimension,
            sample_size,
            empty_size_check=self.empty_size_check,
        )

        if initial_steps + final_steps <= 1:
            return _create_candidates_from(
                empty, space, sample_size, self.inter_bin_randomness
            )

        logging.debug("Creating candidate points to extend the design")
        new_sample = random_search(
            creator=partial(
                _create_candidates_from,
                empty_bins=empty,
                space=space,
                sample_size=sample_size,
                inter_bin_randomness=self.inter_bin_randomness,
            ),
            scorer=scorer,
            steps=initial_steps,
        )
        logging.info("Optimizing candidate points to extend the design")
        return simulated_annealing_by_perturbation(
            new_sample, scorer, steps=final_steps
        )


def create_orthogonal_design(
    space: ParameterSpace,
    sample_size: int,
    inter_bin_randomness: float = 1.0,
) -> np.ndarray:
    """
    Create an orthogonal design without any optimization.

    :meta private:
    """
    # Sometimes, we may randomly generate probabilities with
    # singular correlation matrices. Try 3 times to avoid issue until we give up
    error_text = ""
    for k in range(3):
        probabilities = create_lhd_probabilities(
            space.dimensions, sample_size, inter_bin_randomness=inter_bin_randomness
        )
        doe = space.value_of(probabilities)
        try:
            return iman_connover_transformation(doe, space.correlation)
        except np.linalg.LinAlgError as exc:
            error_text = str(exc)
            pass
    error_text = f"Orthogonal design may not have the desired correlation due to the following error: {error_text}"
    logging.warning(error_text)
    return doe


def create_lhd_probabilities(
    num_variables: int,
    sample_size: int,
    inter_bin_randomness: float = 1.0,
) -> np.ndarray:
    """Create probabilities for a Latin hypercube design.

    :meta private:
    """
    if not 0.0 <= inter_bin_randomness <= 1.0:
        raise ValueError(
            f"inter_bin_randomness has to be between 0 and 1, got {inter_bin_randomness}"
        )
    doe = uniform.rvs(size=(sample_size, num_variables))
    doe = (np.argsort(doe, axis=0) + 0.5) / sample_size
    if inter_bin_randomness == 0.0:
        return doe
    delta = inter_bin_randomness / sample_size
    return doe + uniform(-delta / 2, delta).rvs(size=(sample_size, num_variables))


def _find_sufficient_empty_bins(
    probabilities: np.ndarray,
    bins_per_dimension: int,
    required_sample_size: int,
    empty_size_check: Callable[[np.ndarray], float] = np.max,
) -> np.ndarray:
    empty = _find_empty_bins(probabilities, bins_per_dimension)
    cols = np.where(empty)[1]
    while (
        empty_size_check(np.unique(cols, return_counts=True)[1]) < required_sample_size
    ):
        bins_per_dimension += 1
        empty = _find_empty_bins(probabilities, bins_per_dimension)
        cols = np.where(empty)[1]
    return empty


def _find_empty_bins(probabilities: np.ndarray, bins_per_dimension: int) -> np.ndarray:
    """
    Find empty bins on an orthogonal sampling grid given the probabilities.

    :param probabilities: Array of cdf values of the observed points.
    :param bins_per_dimension: Determines the size of the grid to be tested.
    :return: Boolean array of empty bins with shape=(n_bins, n_dims).
    """
    empty_bins = np.ones((bins_per_dimension, probabilities.shape[1]), dtype=bool)
    edges = np.arange(bins_per_dimension + 1) / bins_per_dimension
    edges = edges.reshape((-1, 1))
    for i_dim in range(probabilities.shape[1]):
        condition = np.logical_and(
            probabilities[:, i_dim] >= edges[:-1], probabilities[:, i_dim] < edges[1:]
        )
        empty_bins[:, i_dim] = np.logical_not(condition.any(1))
    return empty_bins


def _create_candidates_from(
    empty_bins: np.ndarray,
    space: ParameterSpace,
    sample_size: int,
    inter_bin_randomness: float = 1.0,
) -> np.ndarray:
    if not 0.0 <= inter_bin_randomness <= 1.0:
        raise ValueError(
            f"inter_bin_randomness has to be between 0 and 1, got {inter_bin_randomness}"
        )
    empty_rows, empty_cols = np.where(empty_bins)
    bins_per_dimension, dimensions = empty_bins.shape
    delta = 1 / bins_per_dimension
    probabilities = np.empty((sample_size, dimensions))
    for i_dim in range(dimensions):
        values = empty_rows[empty_cols == i_dim]
        np.random.shuffle(values)
        diff = sample_size - values.size
        if diff < 0:
            values = values[:sample_size]
        elif diff > 0:
            available = [idx for idx in range(bins_per_dimension) if idx not in values]
            extra = np.random.choice(available, diff, replace=False)
            values = np.append(extra, values)
        probabilities[:, i_dim] = values * delta + delta / 2
    if inter_bin_randomness > 0.0:
        delta *= inter_bin_randomness
        probabilities += uniform(-delta / 2, delta).rvs(size=(sample_size, dimensions))
    doe = space.value_of(probabilities)
    try:
        return iman_connover_transformation(doe, space.correlation)
    except np.linalg.LinAlgError:
        return doe



import logging
import warnings
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Callable

import numpy as np
from scipy.special import comb as combine

from experiment_design.scorers import Scorer


def random_search(
    creator: Callable[[], np.ndarray], scorer: Scorer, steps: int
) -> np.ndarray:
    """
    Given a |DoE| creator and scorer, maximize the score by random search.

    :param creator: |DoE| creating function.
    :param scorer: |DoE| scoring function.
    :param steps: Number of steps to search.
    :return: The |DoE| matrix with the best score.
    """
    steps = max(1, steps)
    best_doe = creator()
    start_score = best_score = scorer(best_doe)
    logging.debug(f"Initial score: {start_score:.2e}")
    for i_try in range(2, steps + 1):
        doe = creator()
        score = scorer(doe)
        if score > best_score:
            best_doe = doe
            best_score = score
            logging.debug(
                f"Step {i_try} - start score improved by {100 * abs((best_score - start_score) / start_score):.1f}%"
            )
    logging.info(
        f"Final score improved start score by {100 * abs((best_score - start_score) / start_score):.1f}% in {steps} steps"
    )
    return best_doe


def simulated_annealing_by_perturbation(
    doe: np.ndarray,
    scorer: Scorer,
    steps: int = 1000,
    cooling_rate: float = 0.95,
    temperature: float = 25.0,
    max_steps_without_improvement: int = 25,
) -> np.ndarray:
    """
    Simulated annealing algorithm to maximize the score of a |DoE| by
    perturbing the rows of the design matrix along the columns. This kind of perturbation is
    used to avoid violating the |LHS|, i.e. to keep the number of filled bins
    same.

    :param doe: |DoE| matrix with shape (sample_size, len(variables)).
    :param scorer: A scoring function for the doe. It will be maximized.
    :param steps: Number of steps for the annealing algorithm.
    :param cooling_rate: Annealing parameter to decay temperature.
    :param temperature: Annealing temperature.
    :param max_steps_without_improvement: A limit on the maximum steps to take for exploration before setting the
        reference matrix to the last best value.
    :return: Optimized |DoE| matrix

    References
    ----------
    R.V. Joseph and Y. Hung (2008). "`Orthogonal-Maximin Latin Hypercube Designs
    <https://www3.stat.sinica.edu.tw/statistica/oldpdf/A18n17.pdf>`_"

    C. Bogoclu (2022). "`Local Latin Hypercube Refinement for Uncertainty Quantification and Optimization
    <https://hss-opus.ub.ruhr-uni-bochum.de/opus4/frontdoor/deliver/index/docId/9143/file/diss.pdf>`_" Chapter 4.3.2.2

    """
    if temperature <= 1e-16:
        raise ValueError("temperature must be strictly positive.")
    if not 0 <= cooling_rate <= 1:
        raise ValueError("decay has to be between 0 and 1.")

    def cool_down_temperature(temp: float, min_temperature: float = 1e-6) -> float:
        return max(temp * cooling_rate, min_temperature)

    if max_steps_without_improvement < 1:
        max_steps_without_improvement = 1

    doe_start = doe.copy()
    best_doe = doe.copy()
    start_score = anneal_step_score = best_score = scorer(doe)
    logging.debug(f"Initial score: {start_score:.2e}")
    steps_without_improvement = 0
    switch_cache = _MatrixRowSwitchCache(
        row_size=doe.shape[0], column_size=doe.shape[1]
    )
    old_switch_cache = _MatrixRowSwitchCache(
        row_size=doe.shape[0], column_size=doe.shape[1]
    )
    for i_try in range(2, steps + 1):
        if switch_cache.is_full:
            logging.info("No more perturbation left to improve the score")
            break
        doe_try = switch_cache.switch_rows_and_cache(doe_start.copy())
        curr_score = scorer(doe_try)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="overflow")
            transition_probability = np.exp(
                -(anneal_step_score - curr_score) / temperature
            )

        if (
            curr_score > anneal_step_score
            or np.random.random() <= transition_probability
        ):
            doe_start = doe_try.copy()
            old_switch_cache = deepcopy(switch_cache)
            anneal_step_score = curr_score
            switch_cache.reset()
            temperature = cool_down_temperature(temperature)
            if curr_score > best_score:
                best_doe = doe_try.copy()
                best_score = curr_score
                steps_without_improvement = 0
                logging.debug(
                    f"{i_try} - start score improved by {100 * abs((best_score - start_score) / start_score):.1f}%"
                )
        steps_without_improvement += 1

        if steps_without_improvement >= max_steps_without_improvement:
            temperature = cool_down_temperature(temperature)
            # Bound Randomness by setting back to best result
            # This often accelerates convergence speed
            doe_start = best_doe.copy()
            switch_cache = deepcopy(old_switch_cache)
            anneal_step_score = best_score
            steps_without_improvement = 0

    logging.info(
        f"Final score improved start score by {100 * abs((best_score - start_score) / start_score):.1f}% in {steps} steps"
    )
    return best_doe


@dataclass
class _MatrixRowSwitchCache:
    """A cache that is used in simulated annealing to avoid repeating the same perturbations"""

    row_size: int
    column_size: int
    _max_switches_per_column: int = field(init=False, repr=False, default=None)
    _cache: list[np.ndarray] = field(init=False, repr=False, default=None)
    _cache_sizes: np.ndarray = field(init=False, repr=False, default=None)

    def __post_init__(self) -> None:
        # We switch two rows, thus the magic number
        self._max_switches_per_column = combine(self.row_size, 2)
        self.reset()

    @property
    def is_full(self) -> bool:
        return np.min(self._cache_sizes) >= self._max_switches_per_column

    def reset(self) -> None:
        self._cache_sizes = np.zeros(self.column_size, dtype=int)
        self._cache = [np.zeros((0, 2), dtype=int) for _ in range(self.column_size)]

    def choose_column(self) -> int:
        return np.random.choice(
            np.where(self._cache_sizes < self._max_switches_per_column)[0]
        )

    def choose_rows(self, column: int) -> tuple[int, int]:
        def choose_non_occupied(occupied: set[int]) -> int:
            return np.random.choice(
                [idx for idx in range(self.row_size) if idx not in occupied]
            )

        row_cache = self._cache[column]
        max_switches_per_row = self.row_size - 1
        fulls, counts = np.unique(row_cache, return_counts=True)
        row_1 = choose_non_occupied(set(fulls[counts >= max_switches_per_row]))
        row_2 = choose_non_occupied(
            set(row_cache[np.any(row_cache == row_1, axis=1)].ravel())
        )
        return row_1, row_2

    def cache(self, column: int, row_1: int, row_2: int) -> None:
        self._cache[column] = np.append(
            self._cache[column], np.array([[row_1, row_2]]), axis=0
        )
        self._cache_sizes[column] += 1

    def switch_rows_and_cache(self, matrix: np.ndarray) -> np.ndarray:
        column = self.choose_column()
        row_1, row_2 = self.choose_rows(column)
        self.cache(column, row_1, row_2)
        matrix[row_1, column], matrix[row_2, column] = (
            matrix[row_2, column],
            matrix[row_1, column],
        )
        return matrix



from dataclasses import dataclass, field
from typing import Sequence

import numpy as np
from scipy.stats import randint

# Following is ugly, but it is scipy's fault for not exposing rv_frozen
# noinspection PyProtectedMember
from scipy.stats._distn_infrastructure import rv_frozen

from experiment_design import variable
from experiment_design.variable import ContinuousVariable, DiscreteVariable


@dataclass
class ParameterSpace:
    """A container of multiple variables defining a parameter space.

    :param variables: list of variables or marginal distributions that define the marginal parameters
    :param correlation: A float or asymmetric matrix with shape (len(variables), len(variables)), representing the
        linear dependency between the dimensions. If a float is passed, all non-diagonal entries of the unit matrix will
        be set to this value.
    :param infinite_bound_probability_tolerance: If the variable is unbounded, this will be used to extract finite bounds
        as described in lower_bound and upper_bound descriptions. (Default: 1e-6)
    """

    variables: (
        list[
            variable.Variable | variable.ContinuousVariable | variable.DiscreteVariable
        ]
        | list[rv_frozen]
    )
    correlation: float | np.ndarray | None = None
    infinite_bound_probability_tolerance: float = 1e-6

    _lower_bound: np.ndarray = field(init=False, repr=False, default=None)
    _upper_bound: np.ndarray = field(init=False, repr=False, default=None)

    def __post_init__(self) -> None:
        if isinstance(self.variables[0], rv_frozen):
            self.variables = variable.create_variables_from_distributions(
                self.variables
            )

        if self.correlation is None:
            self.correlation = 0
        self.correlation = create_correlation_matrix(self.correlation, self.dimensions)
        if not (
            self.correlation.shape[0] == self.correlation.shape[1] == self.dimensions
        ):
            raise ValueError(
                f"Inconsistent shapes: {self.dimensions} does not match "
                f"{self.correlation.shape}"
            )
        if np.max(np.abs(self.correlation)) > 1:
            raise ValueError("Correlations should be in the interval [-1,1].")

        lower, upper = [], []
        for var in self.variables:
            lower.append(
                var.finite_lower_bound(self.infinite_bound_probability_tolerance)
            )
            upper.append(
                var.finite_upper_bound(self.infinite_bound_probability_tolerance)
            )
        self._lower_bound = np.array(lower)
        self._upper_bound = np.array(upper)

    def _map_by(self, attribute: str, values: np.ndarray) -> np.ndarray:
        if len(values.shape) != 2:
            values = values.reshape((-1, self.dimensions))
        results = np.zeros(values.shape)
        for i_dim, design_variable in enumerate(self.variables):
            results[:, i_dim] = getattr(design_variable, attribute)(values[:, i_dim])
        return results

    def value_of(self, probabilities: np.ndarray) -> np.ndarray:
        """
        Given an array of marginal probabilities, return the corresponding values with shape probabilities.shape using
        the inverse marginal CDF.

        Since it operates on the marginal variables, correlation does not have an effect.
        """
        return self._map_by("value_of", probabilities)

    def cdf_of(self, values: np.ndarray) -> np.ndarray:
        """
        Given an array of marginal values return the marginal probabilities with shape values.shape using the CDF.

        Since it operates on the marginal variables, correlation does not have an effect.
        """
        return self._map_by("cdf_of", values)

    @property
    def lower_bound(self) -> np.ndarray:
        """
        Lower bound values of the space with shape (self.dimensions, ).

        variable.finite_lower_bound is used to provide finite bounds even for unbounded variables
        """
        return self._lower_bound

    @property
    def upper_bound(self) -> np.ndarray:
        """
        Upper bound of the space with shape (self.dimensions, ).

        variable.finite_upper_bound is used to provide finite bounds even for unbounded variables
        """
        return self._upper_bound

    @property
    def dimensions(self) -> int:
        """Size of the space, i.e. the number of variables."""
        return len(self.variables)


VariableCollection = list[rv_frozen] | list[variable.Variable] | ParameterSpace


def create_correlation_matrix(
    target_correlation: float | np.ndarray = 0.0,
    num_variables: int | None = None,
) -> np.ndarray:
    """
    Create a correlation matrix from the target correlation in case it is a float

    :meta private:
    """
    if not np.isscalar(target_correlation):
        return target_correlation
    if not num_variables:
        raise ValueError(
            "num_variables have to be passed if the target_correlation is a scalar."
        )
    return (
        np.eye(num_variables) * (1 - target_correlation)
        + np.ones((num_variables, num_variables)) * target_correlation
    )


def create_discrete_uniform_space(
    discrete_sets: list[list[int | float]],
) -> ParameterSpace:
    """
    Given sets of possible values, create corresponding discrete variables with equal probability of each value.

    :param discrete_sets: list of possible values for each variable
    :return: Parameter space consisting of discrete uniform variables with the same size as the discrete_sets
    """
    variables = []
    for discrete_set in discrete_sets:
        n_values = len(discrete_set)
        if n_values < 2:
            raise ValueError("At least two values are required for discrete variables")
        # In the following, it is OK and even advantageous to have a mutable
        # default argument as a very rare occasion. Therefore, we disable inspection.
        # noinspection PyDefaultArgument
        variables.append(
            DiscreteVariable(
                distribution=randint(0, n_values),
                # Don't forget to bind the discrete_set below either by
                # defining a kwarg as done here, or by generating in another
                # scope, e.g. function. Otherwise, the last value of discrete_sets
                # i.e. the last entry of discrete_sets will be used for all converters
                # Check https://stackoverflow.com/questions/19837486/lambda-in-a-loop
                # for a description as this is expected python behaviour.
                value_mapper=lambda x, values=sorted(discrete_set): values[int(x)],
                inverse_value_mapper=lambda x,
                values=sorted(discrete_set): values.index(x),
            )
        )
    return ParameterSpace(variables)


def create_continuous_uniform_space(
    lower_bounds: Sequence[float], upper_bounds: Sequence[float]
) -> ParameterSpace:
    """
    Given lower and upper bounds, create uniformly distributed variables.

    :param lower_bounds: Array with shape (n_dim,) representing the lower bounds of the uniform variables
    :param upper_bounds: Array with shape (n_dim,) representing the upper bounds of the uniform variables
    :return: Parameter space consisting of continuous uniform variables with the same size as the passed bounds
    """
    if len(lower_bounds) != len(upper_bounds):
        raise ValueError(
            "Number of lower bounds has to be equal to the number of upper bounds"
        )
    variables = []
    for lower, upper in zip(lower_bounds, upper_bounds):
        variables.append(ContinuousVariable(lower_bound=lower, upper_bound=upper))
    return ParameterSpace(variables)



from dataclasses import dataclass
from typing import Any, Callable, Protocol

import numpy as np
from scipy.stats import rv_continuous, rv_discrete, uniform

# Following is ugly, but it is scipy's fault for not exposing rv_frozen
# noinspection PyProtectedMember
from scipy.stats._distn_infrastructure import rv_frozen


def _is_frozen_discrete(dist: Any) -> bool:
    """Check if dist is a rv_frozen_discrete instance"""
    return isinstance(dist, rv_frozen) and isinstance(dist.dist, rv_discrete)


def _is_frozen_continuous(dist: Any) -> bool:
    """Check if dist is a rv_frozen_continuous instance"""
    return isinstance(dist, rv_frozen) and isinstance(dist.dist, rv_continuous)


def _change_field_representation(
    dataclass_instance: dataclass, representations_to_change: dict[str, Any]
) -> str:
    """Just like the default __repr__ but supports reformatting and replacing some values."""
    final = []
    for current_field in dataclass_instance.__dataclass_fields__.values():
        if not current_field.repr:
            continue
        name = current_field.name
        value = representations_to_change.get(
            name, dataclass_instance.__getattribute__(name)
        )
        final.append(f"{name}={value}")
    return f"{dataclass_instance.__class__.__name__}({', '.join(final)})"


def _create_distribution_representation(distribution: rv_frozen) -> str:
    """Create a readable representation of rv_frozen instances"""
    args = ", ".join([str(a) for a in distribution.args])
    kwargs = ", ".join([f"{k}={v}" for k, v in distribution.kwds.items()])
    params = [a for a in [args, kwargs] if a]
    return f"{distribution.dist.name}({', '.join(params)})"


@dataclass
class ContinuousVariable:
    """
    A variable with continuous distribution

    :param distribution: rv_frozen instance representing the distribution. If None (default), it will be set to uniform
        between the passed lower_bound and upper_bound
    :param lower_bound: Lower bound for the variable. If None (default), left support boundary of the distribution will
        be used in case the distribution is bounded. Otherwise, distribution.ppf(infinite_bound_probability_tolerance)
        will be used.
    :param upper_bound: Upper bound for the variable. If None (default), right support boundary of the distribution will
        be used in case the distribution is bounded. Otherwise, distribution.ppf(1 - infinite_bound_probability_tolerance)
        will be used.
    """

    distribution: rv_frozen | None = None
    lower_bound: float | None = None
    upper_bound: float | None = None

    def __post_init__(self) -> None:
        if self.distribution is None and None in [self.lower_bound, self.upper_bound]:
            raise ValueError(
                "Either the distribution or both "
                "lower_bound and upper_bound have to be set."
            )
        if self.distribution is None:
            self.distribution = uniform(
                self.lower_bound, self.upper_bound - self.lower_bound
            )
        if (
            None not in [self.lower_bound, self.upper_bound]
            and self.lower_bound >= self.upper_bound
        ):
            raise ValueError("lower_bound has to be smaller than upper_bound")
        if not _is_frozen_continuous(self.distribution):
            raise ValueError("Only frozen continuous distributions are supported.")

    def value_of(self, probability: float | np.ndarray) -> float | np.ndarray:
        """Given a probability or an array of probabilities return the corresponding value(s) using the inverse |CDF|."""
        values = self.distribution.ppf(probability)
        if self.upper_bound is not None or self.lower_bound is not None:
            return np.clip(values, self.lower_bound, self.upper_bound)
        return values

    def cdf_of(self, value: float | np.ndarray) -> float | np.ndarray:
        """Given a value or an array of values return the probability using the |CDF|."""
        return self.distribution.cdf(value)

    def finite_lower_bound(
        self, infinite_bound_probability_tolerance: float = 1e-6
    ) -> float:
        """
        Provide a finite lower bound of the variable even if it was not provided by the user.

        :param infinite_bound_probability_tolerance: If the variable is unbounded and no explicit lower_bound was
            passed, this will be used to extract finite bounds as described in lower_bound and upper_bound descriptions.
            (Default: 1e-6)
        """
        if self.lower_bound is not None:
            return self.lower_bound
        value = self.value_of(0.0)
        if np.isfinite(value):
            return value
        return self.value_of(infinite_bound_probability_tolerance)

    def finite_upper_bound(
        self, infinite_bound_probability_tolerance: float = 1e-6
    ) -> float:
        """Provide a finite upper bound of the variable even if it was not provided by the user.

        :param infinite_bound_probability_tolerance: If the variable is unbounded and no explicit lower_bound was
            passed, this will be used to extract finite bounds as described in lower_bound and upper_bound descriptions.
            (Default: 1e-6)

        """
        if self.upper_bound is not None:
            return self.upper_bound
        value = self.value_of(1.0)
        if np.isfinite(value):
            return value
        return self.value_of(1 - infinite_bound_probability_tolerance)

    def __repr__(self) -> str:
        distribution_representation = _create_distribution_representation(
            self.distribution
        )
        return _change_field_representation(
            self, {"distribution": distribution_representation}
        )


@dataclass
class DiscreteVariable:
    """
    A variable with discrete distribution

    :param distribution: rv_frozen instance representing the distribution. If None (default), it will be set to uniform between
        the passed lower_bound and upper_bound
    :param value_mapper: Given an integer, i.e. an ordinal encoding, this is expected to return the corresponding
        discrete value of the underlying set of possible values. (Default: lambda x: x)
    :param inverse_value_mapper: Given a discrete value, this is expected to return the corresponding integer value,
        i.e. ordinal encoding. (Default: lambda x: x)
    """

    distribution: rv_frozen
    value_mapper: Callable[[float], float | int] = lambda x: x
    inverse_value_mapper: Callable[[float, int], float] = lambda x: x

    def __post_init__(self) -> None:
        if not _is_frozen_discrete(self.distribution):
            raise ValueError("Only frozen discrete distributions are supported.")
        self.value_mapper = np.vectorize(self.value_mapper)
        self.inverse_value_mapper = np.vectorize(self.inverse_value_mapper)

    def value_of(self, probability: float | np.ndarray) -> float | np.ndarray:
        """Given a probability or an array of probabilities return the corresponding value(s) using the inverse cdf."""
        values = self.distribution.ppf(probability)
        return self.value_mapper(values)

    def cdf_of(self, values: float | np.ndarray) -> float | np.ndarray:
        """Given a value or an array of values return the probability using the cdf."""
        return self.distribution.cdf(self.inverse_value_mapper(values))

    def finite_lower_bound(
        self, infinite_bound_probability_tolerance: float = 1e-6
    ) -> float:
        """
        Provide a finite lower bound of the variable even if it was not provided by the user.

        :param infinite_bound_probability_tolerance: If the variable is unbounded and no explicit lower_bound was
            passed, this will be used to extract finite bounds as described in lower_bound and upper_bound descriptions.
            (Default: 1e-6)
        """
        support = self.distribution.support()
        if np.isfinite(support[0]):
            return self.value_mapper(support[0])
        return self.value_of(infinite_bound_probability_tolerance)

    def finite_upper_bound(
        self, infinite_bound_probability_tolerance: float = 1e-6
    ) -> float:
        """
        Provide a finite upper bound of the variable even if it was not provided by the user.

        :param infinite_bound_probability_tolerance: If the variable is unbounded and no explicit lower_bound was
            passed, this will be used to extract finite bounds as described in lower_bound and upper_bound descriptions.
            (Default: 1e-6)
        """
        support = self.distribution.support()
        if np.isfinite(support[-1]):
            return self.value_mapper(support[1])
        return self.value_of(1 - infinite_bound_probability_tolerance)

    def __repr__(self) -> str:
        distribution_representation = _create_distribution_representation(
            self.distribution
        )
        return _change_field_representation(
            self, {"distribution": distribution_representation}
        )


class Variable(Protocol):
    """A protocol to represent the expected methods of valid Variable objects"""

    @property
    def distribution(self) -> rv_frozen:
        """Distribution of the variable"""

    def value_of(self, probability: float | np.ndarray) -> float | np.ndarray:
        """Given a probability or an array of probabilities return the corresponding value(s) using the inverse cdf."""

    def cdf_of(self, value: float | np.ndarray) -> float | np.ndarray:
        """Given a value or an array of values return the probability using the cdf."""

    def finite_lower_bound(
        self, infinite_bound_probability_tolerance: float = 1e-6
    ) -> float:
        """
        Provide a finite upper bound of the variable even if it was not provided by the user.

        :param infinite_bound_probability_tolerance: If the variable is unbounded and no explicit lower_bound was
            passed, this will be used to extract finite bounds as described in lower_bound and upper_bound descriptions.
            (Default: 1e-6)
        """

    def finite_upper_bound(
        self, infinite_bound_probability_tolerance: float = 1e-6
    ) -> float:
        """
        Provide a finite upper bound of the variable even if it was not provided by the user.

        :param infinite_bound_probability_tolerance: If the variable is unbounded and no explicit lower_bound was
            passed, this will be used to extract finite bounds as described in lower_bound and upper_bound descriptions.
            (Default: 1e-6)
        """


def create_variables_from_distributions(
    distributions: list[rv_frozen],
) -> list[ContinuousVariable | DiscreteVariable]:
    """
    Given a list of distributions, create the corresponding continuous or discrete variables.

    :param distributions: Frozen scipy distributions each representing a marginal variable
    :return: List of variables according to the passed distributions
    """
    variables = []
    for dist in distributions:
        if _is_frozen_discrete(dist):
            variables.append(DiscreteVariable(distribution=dist))
        elif _is_frozen_continuous(dist):
            variables.append(ContinuousVariable(distribution=dist))
        else:
            raise ValueError(
                f"Each distribution must be a frozen discrete or continuous type, got {type(dist)}"
            )
    return variables



import logging
from functools import partial

import numpy as np
from scipy.stats import uniform

from experiment_design.covariance_modification import (
    iman_connover_transformation,
    second_moment_transformation,
)
from experiment_design.experiment_designer import ExperimentDesigner
from experiment_design.optimize import random_search
from experiment_design.scorers import Scorer, ScorerFactory
from experiment_design.variable import ParameterSpace


class RandomSamplingDesigner(ExperimentDesigner):
    """
    Create or extend a |DoE| by randomly sampling from the variable distributions.

    :param exact_correlation: If True, the correlation matrix of the resulting design will match the target correlation
        exactly using a second moment transformation. This may lead variables with finite bounds to generate values that
        are out of bounds. Otherwise, Iman-Connover method will be used, where the values will be kept as is for each
        variable as they are generated from the marginal distribution. This may lead to some imprecision of the
        correlation matrix.
    :param scorer_factory: A factory that creates scorers for the given variables, sample_size and in the cast of an
        extension, old sampling points. If not passed, a default one will be created, that evaluates the maximum
        correlation error and minimum pairwise distance.See
        `experiment_design.scorers.create_default_scorer_factory <#experiment_design.scorers.create_default_scorer_factory>`_
        for more details.

    Examples
    --------
    >>> from experiment_design import create_continuous_uniform_space, RandomSamplingDesigner
    >>> space = create_continuous_uniform_space([-2., -2.], [2., 2.])
    >>> designer = RandomSamplingDesigner()
    >>> doe1 = designer.design(space, sample_size=20)
    >>> doe1.shape
    (20, 2)
    >>> doe2 = designer.design(space, sample_Size=4, old_sample=doe1)
    >>> doe2.shape
    (4, 2)
    """

    def __init__(
        self,
        exact_correlation: bool = False,
        scorer_factory: ScorerFactory | None = None,
    ) -> None:
        self.exact_correlation = exact_correlation
        super(RandomSamplingDesigner, self).__init__(scorer_factory=scorer_factory)

    def _create(
        self,
        space: ParameterSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
    ) -> np.ndarray:
        steps = initial_steps + final_steps
        if steps <= 1:
            return sample_from(space, sample_size, self.exact_correlation)
        return random_search(
            creator=partial(
                sample_from,
                space,
                sample_size,
                self.exact_correlation,
            ),
            scorer=scorer,
            steps=steps,
        )

    def _extend(
        self,
        old_sample: np.ndarray,
        space: ParameterSpace,
        sample_size: int,
        scorer: Scorer,
        initial_steps: int,
        final_steps: int,
    ) -> np.ndarray:
        steps = initial_steps + final_steps
        if steps <= 1:
            return sample_from(space, sample_size, self.exact_correlation)
        logging.warning(
            "If the design space changes, "
            "random sampling may not handle correlation modification properly!"
        )
        return random_search(
            creator=partial(
                sample_from,
                space,
                sample_size,
                self.exact_correlation,
            ),
            scorer=scorer,
            steps=steps,
        )


def sample_from(
    space: ParameterSpace,
    sample_size: int,
    exact_correlation: bool = False,
) -> np.ndarray:
    """
    Sample from the distributions of the variables.

    :param space: Determines the dimensions of the resulting sample.
    :param sample_size: the number of points to be created.
    :param exact_correlation: If True, second moment transformation will be used, which may not respect the finite
    bounds of the marginal distributions. Otherwise, Iman-Connover method will be used, which may yield imprecise
    correlation matrices.
    :return: Sample matrix with shape (len(variables), samples_size).

    :meta private:
    """

    # Sometimes, we may randomly generate probabilities with
    # singular correlation matrices. Try 3 times to avoid issue until we give up
    error_text = ""
    transformer = (
        second_moment_transformation
        if exact_correlation
        else iman_connover_transformation
    )
    for k in range(3):
        doe = uniform(0, 1).rvs((sample_size, space.dimensions))

        doe = space.value_of(doe)

        if (
            np.isclose(space.correlation, np.eye(space.dimensions)).all()
            and not exact_correlation
        ):
            return doe
        try:
            return transformer(doe, space.correlation)
        except np.linalg.LinAlgError as exc:
            error_text = str(exc)
            pass
    raise np.linalg.LinAlgError(error_text)



from experiment_design.variable.variable import (
    ContinuousVariable,
    DiscreteVariable,
    Variable,
    create_variables_from_distributions,
)
from experiment_design.variable.space import (
    ParameterSpace,
    VariableCollection,
    create_correlation_matrix,
    create_discrete_uniform_space,
    create_continuous_uniform_space,
)

__all__ = [
    "ContinuousVariable",
    "DiscreteVariable",
    "Variable",
    "create_variables_from_distributions",
    "ParameterSpace",
    "VariableCollection",
    "create_correlation_matrix",
    "create_discrete_uniform_space",
    "create_continuous_uniform_space",
]
